/*
 * Rules Module
 *
 * Copyright © 1998-2000, 2005 Guillem Jover <guillem@hadrons.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "board.h"
#include "rules.h"

Pattern win_pattern = {0, DIR_H | DIR_V | DIR_D, "****"};

static Pattern pattern[] = {
	/* Attack Alert:
	 *
	 * This patterns does not have to do checks to move the piece
	 * in theory, but if we have three then we must play that.
	*/
	{1, DIR_H | DIR_D, "-***-"},
	{1, DIR_H | DIR_D | DIR_V, "***-"},
	{1, DIR_H | DIR_D, "-***"},
	{1, DIR_H | DIR_D, "**-*"},
	{1, DIR_H | DIR_D, "*-**"},

	/* Defend Alert:
	 *
	 * This patterns have to do checks to move the piece
	 * if we have two we must play the third to force the other
	 * player to play that (distraction).
	 */
	{2, DIR_H | DIR_D, "-**-"},
	{2, DIR_H | DIR_D | DIR_V, "**--"},
	{2, DIR_H | DIR_D, "--**"},
	{2, DIR_H | DIR_D, "*--*"},
	{2, DIR_H | DIR_D, "*-*-"},
	{2, DIR_H | DIR_D, "-*-*"},

	/* Diversion Alert:
	 *
	 * This patterns may be ignored at all if desired.
	 */
	{3, DIR_H | DIR_D | DIR_V, "*---"},
	{3, DIR_H | DIR_D, "-*--"},
	{3, DIR_H | DIR_D, "--*-"},
	{3, DIR_H | DIR_D, "---*"},
};
static count_t pattern_entries = ENTRIES_IN(pattern);

static inline index_t min(index_t a, index_t b)
{
	return (a < b ? a : b);
}

static inline index_t max(index_t a, index_t b)
{
	return (a > b ? a : b);
}

#ifdef DEBUG
#define debug(format, args...) fprintf(stderr, format, ## args)
#else
#define debug(args...)
#endif

/*
 * Allocates & initializes a Line variable
 */
Line *alloc_line(void)
{
	count_t i;
	Line *ln = malloc(sizeof(Line));

	debug("<-- debug mark (alloc_line) -->\n");
	ln->priority = 0;
	ln->x = ln->y = 0;
	ln->dir = 0;
	ln->size = DIM;
	for (i = 0; i < ln->size; i++)
		ln->line[i] = PLAYER_BLANK;
	ln->next = NULL;

	return ln;
}

/*
 * Frees a list LN
 */
void free_line(Line *ln)
{
	Line *tl;

	debug("<-- debug mark (free_line) -->\n");
	while (ln) {
		tl = ln->next;
		free(ln);
		ln = tl;
	}
}

void copy_line(Line *dst, Line *src)
{
	count_t i;

	debug("<-- debug mark (copy_line) -->\n");
	dst->priority = src->priority;
	dst->x = src->x;
	dst->y = src->y;
	dst->dir = src->dir;
	dst->size = src->size;
	for (i = 0; i < src->size; i++)
		dst->line[i] = src->line[i];
	dst->owner = src->owner;
	dst->next = src->next;
}

/*
 * Insert CP in Line LIST
 */
void insert_line(Line **list, Line *cp)
{
	debug("<-- debug mark (insert_line) -->\n");
	cp->next = *list;
	*list = cp;
}

/*
 * Adds a Line ADD at end of LIST
 */
void add_line(Line **list, Line *add)
{
	Line **tl = list;

	debug("<-- debug mark (add_line) -->\n");
	while (*tl)
		tl = &((*tl)->next);
	*tl = add;
}


bool extract_line(index_t line_idx, dir_t dir, Line *ln)
{
	int i, j, delta;

	debug("<-- begin debug mark (extract_line) -->\n");
	if (!ln)
		return false;
	if (!in_range(line_idx))
		return false;

	ln->priority = 1000;
	ln->owner = PLAYER_NONE;
	ln->dir = dir;

	switch (dir) {
	case DIR_H:
		ln->x = 0;
		ln->y = line_idx;
		ln->size = DIM;
		for (i = 0; i < DIM; i++)
			ln->line[i] = board->grid[line_idx][i].who;
		break;
	case DIR_V:
		ln->x = line_idx;
		ln->y = 0;
		ln->size = DIM;
		for (j = 0; j < DIM; j++)
			ln->line[j] = board->grid[j][line_idx].who;
		break;
	case DIR_DP:
		delta = min(HALFDIM, DIM - 1 - line_idx);
		debug(" <-- DP delta %i -->\n", delta);

		ln->x = line_idx + delta;
		ln->y = HALFDIM - delta;
		ln->size = 0;
		debug(" <-- DP ln->y %lu ln->x %lu size %lu -->\n",
		      ln->y, ln->x, ln->size);
		for (j = ln->y, i = ln->x; in_range(j) && in_range(i); j++, i--, ln->size++) {
			debug(" <-- DP j %d i %d size %lu -->\n", j, i, ln->size);
			ln->line[j] = board->grid[j][i].who;
		}
		break;
	case DIR_DS:
		delta = min(HALFDIM, line_idx);

		debug(" <-- DP delta %i -->\n", delta);
		ln->x = line_idx - delta;
		ln->y = HALFDIM - delta;
		ln->size = 0;
		debug(" <-- DS ln->y %lu ln->x %lu size %lu -->\n",
		      ln->y, ln->x, ln->size);
		for (j = ln->y, i = ln->x; in_range(j) && in_range(i); j++, i++, ln->size++) {
			debug(" <-- DS j %d i %d size %lu -->\n", j, i, ln->size);
			ln->line[j] = board->grid[j][i].who;
		}
		break;
	default:
		return false;
	}

	debug("<-- end debug mark (extract_line) -->\n");
	return true;
}

symbol_t equ_pattern(symbol_t sym, Pattern *pat, Line* ln, index_t delta)
{
	index_t i;
	symbol_t symline;

	debug("<-- begin debug mark (equ_pattern) -->\n");
	if (sym != PLAYER_ANY && sym != PLAYER_CROSS && sym != PLAYER_POINT) {
		fprintf(stderr, "<-- debug mark (invalid symbol) -->\n");
		return PLAYER_NONE;
	}
	if (strlen(pat->text) > ln->size) {
		debug("<-- debug mark (invalid size) -->\n");
		return PLAYER_NONE;
	}

	for (i = 0; pat->text[i]; i++) {
		symline = ln->line[i + delta];
		debug("<-- debug mark (symline %c) -->\n", symline);
		switch (pat->text[i]) {
		case PLAYER_NONE:
			debug("<-- debug mark (none selected) -->\n");
			if (symline != PLAYER_BLANK)
				return PLAYER_NONE;
			break;
		case PLAYER_ANY:
			debug("<-- debug mark (any selected) -->\n");
			if (symline == PLAYER_CROSS ||
			    symline == PLAYER_POINT) {
				if (sym == PLAYER_ANY)
					ln->owner = sym = symline;
				else if (sym != symline)
					return PLAYER_NONE;
			} else
				return PLAYER_NONE;
			break;
		default:
			return PLAYER_NONE;
		}
	}

	debug("<-- end debug mark (equ_pattern) -->\n");
	return sym;
}

Line *search_hpat(symbol_t sym, Pattern *pat)
{
	count_t i, j;
	Line test, *ln = NULL;

	debug("<-- begin debug mark (search_hpat) -->\n");
	for (j = 0; j < DIM; j++) {
		if (!extract_line(j, DIR_H, &test))
			continue;
		for (i = 0; i <= test.size - WIN_DIM; i++)
			if (equ_pattern(sym, pat, &test, i) != PLAYER_NONE) {
				Line *tl = alloc_line();
				copy_line(tl, &test);
				insert_line(&ln, tl);
			}
	}

	debug("<-- end debug mark (search_hpat) -->\n");
	return ln;
}

Line *search_vpat(symbol_t sym, Pattern *pat)
{
	count_t i, j;
	Line test, *ln = NULL;

	debug("<-- begin debug mark (search_vpat) -->\n");
	for (i = 0; i < DIM; i++) {
		if (!extract_line(i, DIR_V, &test))
			continue;
		for (j = 0; j <= test.size - WIN_DIM; j++)
			if (equ_pattern(sym, pat, &test, j) != PLAYER_NONE) {
				Line *tl = alloc_line();
				copy_line(tl, &test);
				insert_line(&ln, tl);
			}
	}

	debug("<-- end debug mark (search_vpat) -->\n");
	return ln;
}

Line *search_dpat(symbol_t sym, Pattern *pat)
{
	count_t i, d;
	Line test, *ln = NULL;

	debug("<-- begin debug mark (search_dpat) -->\n");
	for (i = 0; i < DIM; i++) {
		if (!extract_line(i, DIR_DP, &test))
			continue;
		for (d = 0; d <= test.size - WIN_DIM; d++)
			if (equ_pattern(sym, pat, &test, d) != PLAYER_NONE) {
				Line *tl = alloc_line();
				copy_line(tl, &test);
				insert_line(&ln, tl);
			}
		if (!extract_line(i, DIR_DS, &test))
			continue;
		for (d = 0; d <= test.size - WIN_DIM; d++)
			if (equ_pattern(sym, pat, &test, d) != PLAYER_NONE) {
				Line *tl = alloc_line();
				copy_line(tl, &test);
				insert_line(&ln, tl);
			}
	}

	debug("<-- end debug mark (search_dpat) -->\n");
	return ln;
}

Line *search_pattern(symbol_t sym, Pattern *pat)
{
	Line *ln = NULL;

	debug("<-- debug mark (search_pat) -->\n");
	if (pat->valid_dirs & DIR_H)
		add_line(&ln, search_hpat(sym, pat));
	if (pat->valid_dirs & DIR_V)
		add_line(&ln, search_vpat(sym, pat));
	if (pat->valid_dirs & DIR_D)
		add_line(&ln, search_dpat(sym, pat));

	return ln;
}

Move *search_moves(Line *ln)
{
	return NULL;
}

Move *search_traps(Line *ln)
{
	return NULL;
}

bool think(void)
{
	return true;
}

bool computer_play(void)
{
	return true;
}

