/*
 * ncurses output routines
 *
 * User Interface Module
 * Copyright © 2005 Guillem Jover <guillem@hadrons.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <ncurses.h>

/* ncurses happily pollutes the namespace. */
#undef move

#include "player.h"
#include "board.h"
#include "game.h"
#include "ui.h"

static WINDOW *w_board;
static WINDOW *w_choose;
static WINDOW *w_status;

static void draw_hsegment(WINDOW *w, chtype c)
{
	waddch(w, c);
	waddch(w, c);
	waddch(w, c);
}

static void draw_hline(WINDOW *w, chtype l, chtype c, chtype m, chtype r)
{
	int i;

	waddch(w, l);
	for (i = 0; i < DIM - 1; i++) {
		draw_hsegment(w, c);
		waddch(w, m);
	}
	draw_hsegment(w, c);
	waddch(w, r);
}

static bool ncurses_init(int argc, char **argv)
{
	int i;
	chtype *ch;

	if (!initscr()) {
		fprintf(stderr, "error: could not init curses library\n");
		return false;
	}
	if (has_colors())
		start_color();

	cbreak();
	noecho();

	w_choose = newwin(1, 4 * DIM + 1, 1, 2);
	w_board = newwin(2 * DIM + 1, 4 * DIM + 1, 2, 2);
	w_status = newwin(1, COLS - 2, 2 * DIM + 3, 2);

	intrflush(w_choose, FALSE);
	keypad(w_choose, TRUE);

	ch = (chtype *)malloc(sizeof(chtype));
	init_pair(10, COLOR_RED, COLOR_BLACK);
	*ch = COLOR_PAIR(10);
	player->ui_data = (chtype *)ch;

	ch = (chtype *)malloc(sizeof(chtype));
	init_pair(11, COLOR_BLUE, COLOR_BLACK);
	*ch = COLOR_PAIR(11);
	player->other->ui_data = (chtype *)ch;

	draw_hline(w_board, ACS_ULCORNER, ACS_HLINE, ACS_TTEE, ACS_URCORNER);
	for (i = 0; i < DIM - 1; i++) {
		draw_hline(w_board, ACS_VLINE, ' ', ACS_VLINE, ACS_VLINE);
		draw_hline(w_board, ACS_LTEE, ACS_HLINE, ACS_PLUS, ACS_RTEE);
	}
	draw_hline(w_board, ACS_VLINE, ' ', ACS_VLINE, ACS_VLINE);
	draw_hline(w_board, ACS_LLCORNER, ACS_HLINE, ACS_BTEE, ACS_LRCORNER);

	wrefresh(w_board);

	return true;
}

static bool ncurses_done(void)
{
	free(player->ui_data);
	free(player->other->ui_data);

	delwin(w_board);
	delwin(w_choose);
	delwin(w_status);
	endwin();

	return true;
}

static void ncurses_display_board(void)
{
	wrefresh(w_board);
}

static void ncurses_display_winner(void)
{
	mvwprintw(w_status, 0, 0, "player [%c] won ... \n", board->winner);
	wrefresh(w_status);
	sleep(2);
}

static int ncurses_get_user_move(void)
{
	chtype sym = player->symbol | *(chtype *)player->ui_data;
	int key;
	int x = HALFDIM;

	mvwaddch(w_choose, 0, x * 4 + 2, sym);
	wrefresh(w_choose);

	do {
		key = wgetch(w_choose);

		switch (key) {
		case 'l':
		case KEY_RIGHT:
			mvwaddch(w_choose, 0, x * 4 + 2, ' ');
			if (x < DIM - 1)
				x++;
			break;
		case 'h':
		case KEY_LEFT:
			mvwaddch(w_choose, 0, x * 4 + 2, ' ');
			if (x > 0)
				x--;
			break;
		case 'q':
		case 'Q':
		case KEY_EXIT:
			ncurses_done();
			exit(0);
			break;
		}

		mvwaddch(w_choose, 0, x * 4 + 2, sym);
		wrefresh(w_choose);
	} while (key != '\n');

	mvwaddch(w_choose, 0, x * 4 + 2, ' ');
	wrefresh(w_choose);

	return x;
}

static void ncurses_draw_board_move(int x, int y)
{
	chtype sym = player->symbol | *(chtype *)player->ui_data;

	mvwaddch(w_board, (DIM - y - 1) * 2 + 1, x * 4 + 2, sym);
}

struct ui_module ui_module = {
	.name = "ncurses",
	.init = ncurses_init,
	.done = ncurses_done,
	.display_board = ncurses_display_board,
	.display_winner = ncurses_display_winner,
	.get_user_move = ncurses_get_user_move,
	.draw_board_move = ncurses_draw_board_move,
};

